#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <stdbool.h>
#include <ncurses.h>
#include <unistd.h>
#include <termios.h>
#include <limits.h>
#include <time.h>
#include <SDL2/SDL.h>
#include <SDL2/SDL_image.h>

#include "hash.h"
#include "structure.h"
#include "base.h"
#include "move.h"
#include "generation.h"
#include "display.h"

void updateRenderer(SDL_Renderer* renderer) {
    //printf("E");
    SDL_RenderPresent(renderer);
}

void resetRenderer(SDL_Renderer* renderer) {
    SDL_SetRenderDrawColor(renderer, 0, 0, 0, 255);
    SDL_RenderClear(renderer);
}

void drawRectToRenderer(SDL_Renderer* renderer, SDL_Rect* rect, int R, int G, int B, int A) {
    SDL_SetRenderDrawColor(renderer, R, G, B, A);
    SDL_RenderFillRect(renderer, rect);
}

void placeRectToRenderer(SDL_Renderer* renderer, int X, int Y, int W, int H, int R, int G, int B, int A) {
    SDL_Rect rect;
    rect.x = X;
    rect.y = Y;
    rect.w = W;
    rect.h = H;
    SDL_SetRenderDrawColor(renderer, R, G, B, A);
    SDL_RenderFillRect(renderer, &rect);
}

void placeRectToRendererNoColor(SDL_Renderer* renderer, int X, int Y, int W, int H) {
    SDL_Rect rect;
    rect.x = X;
    rect.y = Y;
    rect.w = W;
    rect.h = H;
    SDL_RenderFillRect(renderer, &rect);
}

void drawLineWithThicc(SDL_Renderer* renderer, int width, int x1, int x2, int y1, int y2, int R, int G, int B, int A) {
    // draws line with width (native SDL cannot do that)
    double theta = 0.0;
    double seglen = distance_pt_pt_3d(x1, y1, 0.0, x2, y2, 0.0);
    while(theta < 1.0) {
        placeRectToRenderer(renderer, convex_seg(x1, x2, theta)-width/2, convex_seg(y1, y2, theta)-width/2, width, width, R, G, B, A);
        theta += 1 / seglen;
    }
}

void drawLineWithThiccNoColor(SDL_Renderer* renderer, int width, int x1, int x2, int y1, int y2) {
    // draws line with width (native SDL cannot do that)
    double theta = 0.0;
    double seglen = distance_pt_pt_3d(x1, y1, 0.0, x2, y2, 0.0);
    while(theta < 1.0) {
        placeRectToRendererNoColor(renderer, convex_seg(x1, x2, theta)-width/2, convex_seg(y1, y2, theta)-width/2, width, width);
        theta += 1 / seglen;
    }
}

void drawLineWithThiccGradient(SDL_Renderer* renderer, int start_width, int end_width, int x1, int x2, int y1, int y2, int R, int G, int B, int A) {
    // draws line with width ;
    // width goes from start_width to end_with in a linear way
    double theta = 0.0;
    double seglen = distance_pt_pt_3d(x1, y1, 0.0, x2, y2, 0.0);
    while(theta < 1.0) {
        int width = convex_seg(start_width, end_width, theta) ;
        placeRectToRenderer(renderer, convex_seg(x1, x2, theta)-width/2, convex_seg(y1, y2, theta)-width/2, width, width, R, G, B, A);
        theta += 1 / seglen;
    }
}

void drawDigitToRenderer(SDL_Renderer* renderer, imgs data, int digit, int X, int Y, int W, int H) {
    if(digit == -727) {
        SDL_Rect rect;
        rect.x = X;
        rect.y = Y;
        rect.w = W;
        rect.h = H;
            
        SDL_Texture* texture = data.arr[10];
        
        SDL_RenderCopy(renderer, texture, NULL, &rect);
    } else if (!(0 <= digit && digit <= 9)) {
        fprintf(stderr, "Illegal digit : '%d'.\n", digit);
        exit(1);
    } else {
        SDL_Rect rect;
        rect.x = X;
        rect.y = Y;
        rect.w = W;
        rect.h = H;
            
        SDL_Texture* texture = data.arr[digit];
        
        SDL_RenderCopy(renderer, texture, NULL, &rect);
    }
}

void drawCharToRenderer(SDL_Renderer* renderer, imgs data, char c, int X, int Y, int W, int H) {
    if ((int)c >= 97 && (int)c <= 122) {
        SDL_Rect rect;
        rect.x = X;
        rect.y = Y;
        rect.w = W;
        rect.h = H;
            
        SDL_Texture* texture = data.arr[(int)c - 97];
        
        SDL_RenderCopy(renderer, texture, NULL, &rect);
    }
}

void drawNumberToRenderer(SDL_Renderer* renderer, imgs data, int n, int X, int Y, int W, int H, int Woffset) {
    if(n == 0) {
        drawDigitToRenderer(renderer, data, 0, X + W, Y, W, H);
    } else if(n > 0) {
        int toDraw = 0, remaining = n, nIter = ln_baseN(n, 10);
        while(nIter != 0) {
            toDraw = remaining%10;
            remaining = remaining / 10;
            drawDigitToRenderer(renderer, data, toDraw, X + (W-Woffset)*nIter, Y, W, H);
            nIter--;
        }
    } else {
        int toDraw = 0, remaining = -n, nIter = ln_baseN(-n, 10);
        drawDigitToRenderer(renderer, data, -727, X, Y, W, H);
        while(nIter != 0) {
            toDraw = remaining%10;
            remaining = remaining / 10;
            drawDigitToRenderer(renderer, data, toDraw, X + (W-Woffset)*nIter, Y, W, H);
            nIter--;
        }
    }
}

void drawStringToRenderer(SDL_Renderer* renderer, imgs data, char* s, int X, int Y, int W, int H) {
    int k = 0 ;
    while(s[k] != '\0') {
        drawCharToRenderer(renderer, data, s[k], X + W*k, Y, W, H);
        k += 1;
    }
}

// ---------------------------------------------------------------------------------------------------------------------------------- //
// 3D stuff

void project_to_camera(double x0, double y0, double z0, double* rx, double* ry, double* rz) {
    // align camera to (0, 0, 0)
    double x = x0 - camx ;
    double y = y0 - camy ;
    double z = z0 - camz ;

    // rotate (y)
    double xry = x*cos(rot_hz) + z*sin(rot_hz) ;
    double yry = y ;
    double zry = z*cos(rot_hz) - x*sin(rot_hz) ;

    // rotate (x)
    *rx = xry ;
    *ry = yry*cos(rot_vt) - zry*sin(rot_vt) ;
    *rz = zry*cos(rot_vt) + yry*sin(rot_vt) ;
}

void draw_segment(SDL_Renderer* renderer, double sx, double sy, double sz, double ex, double ey, double ez) {
    double spx ;double spy ;double spz ;double epx ;double epy ;double epz ;
    project_to_camera(sx, sy, sz, &spx, &spy, &spz) ;
    project_to_camera(ex, ey, ez, &epx, &epy, &epz) ;
    if(spz >= 0.4 && epz >= 0.4) {
        drawLineWithThiccNoColor(renderer, 1, 
            (int)(1500.0 * (1.0 + (spx / (1.5 * spz * tan_fov))) / 2.0),
            (int)(1500.0 * (1.0 + (epx / (1.5 * epz * tan_fov))) / 2.0),
            (int)(1000.0 * (1.0 + (spy / (spz * tan_fov))) / 2.0),
            (int)(1000.0 * (1.0 + (epy / (epz * tan_fov))) / 2.0)
        ) ;
    } else if(epz >= 0.4) {
        double midx = convex_pt(ex, sx, (epz - 0.4)/(epz - spz)) ;
        double midy = convex_pt(ey, sy, (epz - 0.4)/(epz - spz)) ;
        double midz = convex_pt(ez, sz, (epz - 0.4)/(epz - spz)) ;
        project_to_camera(midx, midy, midz, &spx, &spy, &spz) ;
        //printf("* %f\n", spz) ;
        drawLineWithThiccNoColor(renderer, 1, 
            (int)(1500.0 * (1.0 + (spx / (1.5 * spz * tan_fov))) / 2.0),
            (int)(1500.0 * (1.0 + (epx / (1.5 * epz * tan_fov))) / 2.0),
            (int)(1000.0 * (1.0 + (spy / (spz * tan_fov))) / 2.0),
            (int)(1000.0 * (1.0 + (epy / (epz * tan_fov))) / 2.0)
        ) ;
    } else if(spz >= 0.4) {
        double midx = convex_pt(sx, ex, (spz - 0.4)/(spz - epz)) ;
        double midy = convex_pt(sy, ey, (spz - 0.4)/(spz - epz)) ;
        double midz = convex_pt(sz, ez, (spz - 0.4)/(spz - epz)) ;
        project_to_camera(midx, midy, midz, &epx, &epy, &epz) ;
        //printf("%f *\n", epz) ;
        drawLineWithThiccNoColor(renderer, 1, 
            (int)(1500.0 * (1.0 + (spx / (1.5 * spz * tan_fov))) / 2.0),
            (int)(1500.0 * (1.0 + (epx / (1.5 * epz * tan_fov))) / 2.0),
            (int)(1000.0 * (1.0 + (spy / (spz * tan_fov))) / 2.0),
            (int)(1000.0 * (1.0 + (epy / (epz * tan_fov))) / 2.0)
        ) ;
    } // else : not visible (behind camera)
}

void axialRotation_X0(double* y, double* z, double theta) {
    double y0 = *y ;
    *y = (*y)*cos(theta) - (*z)*sin(theta) ;
    *z = (*z)*cos(theta) + y0*sin(theta) ;
}

void axialRotation_X(double* y, double* z, double theta, double cst_y, double cst_z) {
    // project the space onto the (y, z) plane to get cst_y and cst_z
    double y1 = *y - cst_y;
    double z1 = *z - cst_z;
    axialRotation_X0(&y1, &z1, theta);
    *y = y1 + cst_y ;
    *z = z1 + cst_z ;
}

void axialRotation_Y0(double* x, double* z, double theta) {
    double x0 = *x ;
    *x = (*x)*cos(theta) + (*z)*sin(theta) ;
    *z = (*z)*cos(theta) - x0*sin(theta) ;
}

void axialRotation_Y(double* x, double* z, double theta, double cst_x, double cst_z) {
    // project the space onto the (y, z) plane to get cst_y and cst_z
    double x1 = *x - cst_x;
    double z1 = *z - cst_z;
    axialRotation_Y0(&x1, &z1, theta);
    *x = x1 + cst_x ;
    *z = z1 + cst_z ;
}

void axialRotation_Z0(double* x, double* y, double theta) {
    double x0 = *x ;
    *x = (*x)*cos(theta) - (*y)*sin(theta) ;
    *y = (*y)*cos(theta) + x0*sin(theta) ;
}

void axialRotation_Z(double* x, double* y, double theta, double cst_x, double cst_y) {
    // project the space onto the (y, z) plane to get cst_y and cst_z
    double x1 = *x - cst_x;
    double y1 = *y - cst_y;
    axialRotation_Z0(&x1, &y1, theta);
    *x = x1 + cst_x ;
    *y = y1 + cst_y ;
}

void drawSegmentRotated(SDL_Renderer* renderer, double sx, double sy, double sz, double ex, double ey, double ez, double hz_angle, double vt_angle, double center_x, double center_y, double center_z) {
    double psx = sx;
    double psy = sy;
    double psz = sz;
    double pex = ex;
    double pey = ey;
    double pez = ez;
    axialRotation_Y(&psx, &psz, hz_angle, center_x, center_z);
    axialRotation_Y(&pex, &pez, hz_angle, center_x, center_z);
    axialRotation_X(&psy, &psz, vt_angle, center_y, center_z);
    axialRotation_X(&pey, &pez, vt_angle, center_y, center_z);
    draw_segment(renderer, psx, psy, psz, pex, pey, pez);
}

void drawOutlineOfCube_0(SDL_Renderer* renderer, cube_0 c) {
    SDL_SetRenderDrawColor(renderer, c.red, c.green, c.blue, 255) ;
    // x = constant
    drawSegmentRotated(renderer, c.x, c.y, c.z, c.x + c.w, c.y, c.z, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;
    drawSegmentRotated(renderer, c.x, c.y + c.h, c.z, c.x + c.w, c.y + c.h, c.z, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;
    drawSegmentRotated(renderer, c.x, c.y, c.z + c.d, c.x + c.w, c.y, c.z + c.d, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;
    drawSegmentRotated(renderer, c.x, c.y + c.h, c.z + c.d, c.x + c.w, c.y + c.h, c.z + c.d, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;

    // y = constant
    drawSegmentRotated(renderer, c.x, c.y, c.z, c.x, c.y + c.h, c.z, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;
    drawSegmentRotated(renderer, c.x + c.w, c.y, c.z, c.x + c.w, c.y + c.h, c.z, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;
    drawSegmentRotated(renderer, c.x, c.y, c.z + c.d, c.x, c.y + c.h, c.z + c.d, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;
    drawSegmentRotated(renderer, c.x + c.w, c.y, c.z + c.d, c.x + c.w, c.y + c.h, c.z + c.d, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;

    // z = constant
    drawSegmentRotated(renderer, c.x, c.y, c.z, c.x, c.y, c.z + c.d, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;
    drawSegmentRotated(renderer, c.x + c.w, c.y, c.z, c.x + c.w, c.y, c.z + c.d, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;
    drawSegmentRotated(renderer, c.x, c.y + c.h, c.z, c.x, c.y + c.h, c.z + c.d, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;
    drawSegmentRotated(renderer, c.x + c.w, c.y + c.h, c.z, c.x + c.w, c.y + c.h, c.z + c.d, c.hz_angle, c.vt_angle, c.x + c.w/2.0, c.y + c.h/2.0, c.z + c.d/2.0) ;
}

void drawCurrentRoom(SDL_Renderer* renderer) {
    for(int k = 0; k < current_room->map_size; k++) {
        drawOutlineOfCube_0(renderer, current_room->map[k]);
    }
    for(int k = 0; k < current_room->tps_size; k++) {
        drawOutlineOfCube_0(renderer, current_room->tps[k].hitbox);
    }
}

// -------------------------------------------------------------------------------------------------------------------------------- //

void drawData(SDL_Renderer* renderer) {
    drawNumberToRenderer(renderer, digits, (int)camx, 10, 10, 75/2, 105/2, 0) ;
    drawNumberToRenderer(renderer, digits, (int)camy, 10, 60, 75/2, 105/2, 0) ;
    drawNumberToRenderer(renderer, digits, (int)camz, 10, 110, 75/2, 105/2, 0) ;

    drawNumberToRenderer(renderer, digits, (int)(rot_hz*180.0/3.14159), 10, 160, 75/2, 105/2, 0) ;
    drawNumberToRenderer(renderer, digits, (int)(rot_vt*180.0/3.14159), 10, 210, 75/2, 105/2, 0) ;

    drawNumberToRenderer(renderer, digits, player_chx, 310, 10, 75/2, 105/2, 0);
    drawNumberToRenderer(renderer, digits, player_chy, 310, 60, 75/2, 105/2, 0);
}